import shutil
import os, sys
import xml.dom.minidom
import re
from boolat_lib import *
import platform
"""
INSTRUCTION:
    Description:
        The system copies files and folders recursively from src path to destination. It supports ignoring some part of
        resources with regular expressions. You can also specify exceptions to ignore expressions to better specify the
        rules.
        If destination already exists, the system will try to use some of existing resources in order to reduce amount
        of work. The system will replace those files and folders that match criterias from boolat_lib.py::need_refresh(src,dest).
        (The destination resource will be replaced if the source was last modified later that destination or the size of
        files is different.)
        At the same time if there are some resources in destination that doesn't exist in source, they will be removed.
        The main rule: destination will be the exact copy of source excluding ignored resources. And destination will
        never contain anything that doesn't exist in source.
    Arguments
        src  - source path
        dest - destination path
        xml_filepath - path to configuration path that provides ignore settings. The system can't work without this file
        to_lower_case - all folder and file names will be lovercased if it is set
    XML
        Config xml Sample:
            <config>
                <_common>
                    <ignore mask="\.svn" recursive="1"/>
                    <ignore mask="\.git" recursive="1"/>
                    <ignore mask="\.DS_Store" recursive="1"/>
                    <ignore mask="thumbs\.db" recursive="1"/>
                </_common>
                <tex>
                    <ignore root="scenes" mask="paywall" recursive="1"/>
                    <exclusion root="scenes/main_menu" mask="paywall"/>
                </tex>
                <assets>
                    <ignore root="scenes" mask=".survey." recursive="1"/>
                </assets>
            </config>
        Xml file is cofigured to support more than one copying source. Is is useful if you copy a set of folders. In this
        case you call smart_copy for each source and pass a path to the same xml file. Each time a set of rules will be
        collected from xml file. It will contain rules from xml node '_common' and from xml node with the same name as
        src folder name. i. e.: if you copy tex folder the set of rules will contain 4 rules from the '_common' node and
        2 rules from 'tex' node.
        Firstly each copying resource is checked for all ignores and if it matches at least one ignore rule, the exclusion
        check takes place. The system checks the resource for all exclusions and if at least one matches, the resource won't
        be ignored and the system will copy it.
        All ignore or exclusion rules are served in the same way and consist of:
            recursive - if it is set to '1' or 'true' the rule will be applied for the folder specified in root and for
                all the subfolders recursively
            root - path to a subfolder to whose content the system will apply the mask.
                root string must start right with subfolder name. For instance: if you copy folder 'tex' and you want to
                specify rule with root 'tex/scenes/main_menu' you must type root='scenes/main_menu'. However these will
                also work: './scenes'; '/scenes' or '.scenes'
                If root is absent it is used as '.' and the mask is applied only for entities in root folder. Or for all
                resources if recursive flag is set. For example: you copy folder 'tex' and have rule with mask="\.DS_Store"
                and recursive="1". In this case all files named ".DS_Store" will be ignored in tex folder and in all subfolders
            mask - regular expression that will be applied for each entity contained by root folder.
                Characters '^' and '$' will be added to the beginning and to the end of the mask string if absent
                Mask is applied for file or folder name only. Not for entire path.
"""


def smart_copy(src, dest, xml_filepath, to_lower_case):
    ignores = []
    if xml_filepath:
        print xml_filepath
        if not os.path.exists(xml_filepath):
            print "ERROR(smart_copy): can't find copy config xml '%s" % xml_filepath
            return 1, []
        xml_doc = xml.dom.minidom.parse(xml_filepath)
        if 0 != smart_copy_entries_from_nodes(ignores, getXmlNodes_FromNode(xml_doc, "config/rule")):
            return 1, ['failed to read config xml']
        if not ignores:
            return 1, ['failed to read config xml: no ignores for this copying source found']
    else:
        print "copying without configuration"
    rules = {'ignores': ignores}
    should_print_all_rules = True
    if should_print_all_rules:
        print "IGNORES"
        for item in ignores:
                print "Root: '%s';    Match: '%s'    Recursive: '%d'" % (item['root'], item['match'], item['recursive'])
        print "-----------"
    if os.path.isdir(src):
        return smart_copy_implementation(src, dest, rules, to_lower_case)
    else:
        # copy file
        output_strings = []
        if not should_ignore(src, os.path.split(src)[1], rules):
            if os.path.exists(dest) and os.path.isdir(dest):
                output_strings.append("remove '%s'" % dest)
                universal_remove(dest)
            if need_refresh(src, dest):
                mk_dir_for_filepath(dest)
                output_strings.append("copy '%s'" % src)
                shutil.copy2(src, dest)
        return 0, output_strings


def smart_copy_implementation(src, dest, rules, to_lower_case):
    if not os.path.exists(src):
        print "ERROR(smart_copy): '%s' not a valid path" % src
        return 1, []
    if not os.path.isdir(src):
        print "ERROR(smart_copy): '%s' must be path to a directory" % src
        return 1, []

    output_strings = []
    # making current rules
    cur_rules = {'ignores': current_rules(rules['ignores'], src)}

    if os.path.exists(dest) and not os.path.isdir(dest):
        output_strings.append("remove '%s'" % dest)
        universal_remove(dest)
    if not os.path.exists(dest):
        output_strings.append("make dir '%s'" % dest)
        mk_dir_recursively(dest)
    src_items = os.listdir(src)
    src_items = filter(lambda x: (not should_ignore(src, x, cur_rules)), src_items)  # apply rules to the src_list here
    if platform.system() == 'Windows':
        to_lower_case=True
    if platform.system() == 'Linux':
        to_lower_case=False
    if to_lower_case:
        src_items = map(lambda x: x.lower(), src_items)
    # remove all dest content if there is no match in src_items
    dest_items = os.listdir(dest)
    for i in dest_items:
        if not i in src_items:
            path = os.path.join(dest, i)
            output_strings.append("remove '%s'" % path)
            universal_remove(path)

    # copy
    for i in src_items:
        src_path = os.path.join(src, i)
        dest_path = os.path.join(dest, i)
        if os.path.isdir(src_path):
            # copy dir
            call_retval, call_out = smart_copy_implementation(src_path, dest_path, cur_rules, to_lower_case)
            if call_retval:
                output_strings.extend(call_out)
                return call_retval, output_strings
            output_strings.extend(call_out)
        else:
            # copy file
            if os.path.exists(dest_path) and os.path.isdir(dest_path):
                output_strings.append("remove '%s'" % dest_path)
                universal_remove(dest_path)
            if need_refresh(src_path, dest_path):
                mk_dir_for_filepath(dest_path)
                output_strings.append("copy '%s'" % src_path)
                shutil.copy2(src_path, dest_path)
    return (0, output_strings)


def should_ignore(src, src_name, rules):
    ignoring = False
    for item in rules['ignores']:
        if not item['root']:
            if re.match(item['match'], src_name, re.I):
                ignoring = True
                break
    if ignoring:
        print "Ignored: " + src + os.path.sep + src_name
    return ignoring


def current_rules(items, src):
    temp, src_name = os.path.split(src)
    out_list = []
    for item in items:
        should_add = False
        root = item['root']
        if root:
            if root.startswith("."):
                root = root[1:]
                should_add = True
            else:
                if root.startswith(src_name):
                    root = root[len(src_name) + 1:]
                    should_add = True
                else:
                    should_add = False
        else:
            should_add = item['recursive']
        if should_add:
            out_list.append({'match': item['match'], 'root': root, 'recursive': item['recursive']})
    return out_list


def smart_copy_entries_from_nodes(out_list, items):
    for item in items:
        match = str(item.getAttribute("match"))
        if not match:
            print "ERROR: smart_copy config file has entry without match"
            return 1
        if match[0] == '*':
            match = '.' + match
        if match[0] != '^':
            match = '^' + match
        if not match.endswith('$'):
            match = match + '$'

        recursive = False
        rec_str = str(item.getAttribute("recursive"))
        if rec_str == '1':
            recursive = True
        root = str(item.getAttribute("root"))
        root = os.path.normpath(root)
        while root[0] == '/' or root[0] == '\\':
            root = root[1:]
        if root[0] != '.':
            root = '.' + root
        out_list.append({'match': match, 'root': root, 'recursive': recursive})
    return 0

#=======================================================================================================================

'''def main():
    smart_copy(src, dest, xml_filepath)
main()'''
